math.randomseed(os.time())
nextmode = {top = 'perspective', perspective = 'top'}
extra_nextmode = {top = 'perspective', perspective = 'perspective1', perspective1 = 'perspective2', perspective2 = 'perspective3', perspective3 = 'top'}
player = {x = 0, y = 0, angle = 0, direction = nil, dest = 0}
goal = {x = 0, y = 0}
speed = 2
function love.load()
	genmaze()
	drawmode = 'top' -- 'perspective'
	mode = 'stopped'
	love.graphics.setBackgroundColor(255,255,255)
	love.graphics.setColor(0,0,0)
	love.graphics.setLineWidth(2)
end

local rnd = math.random
function rnddir(x, y)
	local chance = rnd()
	if chance < .05 then
		local cx, cy
		if goal.x < x then
			cx = -1
		elseif goal.x > x then
			cx = 1
		end
		if goal.y < y then
			cy = -1
		elseif goal.y > y then
			cy = 1
		end
		if cx and cy then
			if rnd() < .5 then
				cx = 0
			else
				cy = 0
			end
		end
		return x + (cx or 0), y + (cy or 0)
	else
		local c = math.random(4)
		return x + (c < 4 and c - 2 or 0), y + (c > 1 and c - 3 or 0)
	end
end

function genmaze()
	player.x = 0
	player.y = 0
	goal.x = 10
	goal.y = 10
	maze = {{x=0, y=0, conn={}}}
	local t = maze[1]
	pointstaken = {['0/0']=t}
	local ptext = goal.x .. '/'..goal.y
	while #maze < 100 or not pointstaken[ptext] do
		local nx, ny = rnddir(t.x, t.y)
		local ntext = nx..'/'..ny
		if not pointstaken[ntext] then
			local nt = {x=nx, y=ny, conn={t}, parent=t}
			pointstaken[ntext] = nt
			table.insert(maze, nt)
			table.insert(t.conn, nt)
			if rnd() < .6 then
				t = nt
			else
				t = t.conn[rnd(#t.conn)]
			end
		end
		if math.random() < 0.05 then
			t = maze[math.random(#maze)]
		end
		while t.x == goal.x and t.y == goal.y do
			t = maze[math.random(#maze)]
		end
	end
end

local function isconnected(conn, x, y)
	for _,t in ipairs(conn) do
		if t.x == x and t.y == y then
			return true
		end
	end
end

local function getdx(angle)
	return angle%2==0 and 0 or 2 - angle
end

local function getdy(angle)
	return angle%2==1 and 0 or 1 - angle
end

function love.update(dt)
	if player.x == goal.x and player.y == goal.y then
		genmaze()
		mode = 'stopped'
		return
	end
	if mode == 'stopped' then
		local down = love.keyboard.isDown
		if down'w' or down'up' and
			isconnected(pointstaken[player.x..'/'..player.y].conn, player.x + getdx(player.angle), player.y + getdy(player.angle)) then
				mode = 'moving'
				player.tomove = 1
				player.dest = {x = player.x + getdx(player.angle), y = player.y + getdy(player.angle)}
			--end
		elseif down'a' or down'left' then
			mode = 'turning'
			player.tomove = 2
			player.dest = player.angle - 1
			player.direction = -1
		elseif down'd' or down'right' then
			mode = 'turning'
			player.tomove = 2
			player.dest = player.angle + 1
			player.direction = 1
		end
	else
		player.tomove = player.tomove - dt*speed
		if player.tomove < 0 then
			if mode == 'moving' then
				player.x = player.dest.x
				player.y = player.dest.y
			end
			mode = 'stopped'
			return
		end
		if mode == 'turning' then
			player.angle = player.angle + dt * player.direction*speed
			if player.direction > 0 then
				if player.angle > player.dest then
					player.angle = player.dest % 4
					mode = 'stopped'
				end
			else
				if player.angle < player.dest then
					player.angle = player.dest % 4
					mode = 'stopped'
				end
			end
			-- player.angle = player.angle % 4
		elseif mode == 'moving' then
			player.x = player.x + getdx(player.angle)*dt*speed
			player.y = player.y + getdy(player.angle)*dt*speed
		end
	end
end

function transform(x, y, ax, ay)
	if drawmode == 'top' then
		x, y = (x-player.x)*30, (y-player.y)*30
		x, y = x*ax - y*ay, y*ax + x*ay
		return x + 400, 300 - y
	elseif drawmode == 'perspective2' then
		x, y = (x-player.x)*30, (y-player.y)*30
		x, y = x*ax - y*ay, y*ax + x*ay
		return x * (300 - y) / 300  + 400, 300 - y + (0.03*y)^2
	elseif drawmode == 'perspective1' then
		x, y = (x-player.x)*30, (y-player.y)*30
		x, y = x*ax - y*ay, y*ax + x*ay
		return x * (300 - y) / 300 + 400, 300 - y
	elseif drawmode == 'perspective3' then
		x, y = (x-player.x)*30, (y-player.y)*30
		x, y = x*ax - y*ay, y*ax + x*ay
		local d = math.sqrt(x*x + y*y + 200*200) -- camera distance
		local a = 250/d -- magnification
		return x * a  + 400, 300 - y * a
	elseif drawmode == 'perspective' then -- this part is where hell breaks loose
		local cos, sin = math.cos, math.sin
		local a_x, a_y, a_z = x*30, y*30, 0
		local t_x, t_y, t_z = ax, ay, 0
		local c_x, c_y, c_z = player.x*30, player.y*30, 0
		local d_x = cos(t_y)*(sin(t_z)*(a_y-c_y)+cos(t_z)*(a_x-c_x))-sin(t_y)*(a_z-c_z)
		local d_y = sin(t_x)*(cos(t_y)*(a_z-c_z)+sin(t_y)*(sin(t_z)*(a_y-c_y)+cos(t_z)*(a_x-c_x)))+cos(t_x)*(cos(t_z)*(a_y-c_y)-sin(t_z)*(a_x-c_x))
		local d_z = cos(t_x)*(cos(t_y)*(a_z-c_z)+sin(t_y)*(sin(t_z)*(a_y-c_y)+cos(t_z)*(a_x-c_x)))-sin(t_x)*(cos(t_z)*(a_y-c_y)-sin(t_z)*(a_x-c_x))
		local e_x, e_y, e_z = player.x*30, player.y*30, 100
		local q = e_z / d_z
		return (d_x - e_x)/q + 400, 300 - (d_y - e_y)/q
	end
end

function drawline(x1, y1, x2, y2, ax, ay)
	x1, y1 = transform(x1, y1, ax, ay)
	x2, y2 = transform(x2, y2, ax, ay)
	love.graphics.line(x1, y1, x2, y2)
end

local cos, sin, hpi = math.cos, math.sin, .5*math.pi
function love.draw()
	local a = player.angle*hpi
	local ax, ay = cos(a), sin(a)
	for i, v in ipairs(maze) do
		if v.parent then
			drawline(v.x, v.y, v.parent.x, v.parent.y, ax, ay)
		end
	end
	love.graphics.circle('fill', 400, 300, 6, 40)
	love.graphics.circle('line', 400, 300, 6, 40)
	local x,y = transform(goal.x, goal.y, ax, ay)
	love.graphics.circle('line', x, y, 6, 40)
end

function love.keypressed(key)
	if key == 'tab' then
		drawmode = nextmode[drawmode] or 'top'
	elseif key == 'f1' then
		nextmode, extra_nextmode = extra_nextmode, nextmode
	elseif key == 'r' then
		genmaze()
	elseif key == 'escape' then
		love.event.push'q'
	end
end
